﻿using System;

using NATS.Client;

namespace TestCommon
{
    /// <summary>
    /// 수신기
    /// </summary>
    public class Receiver
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 메시지 수신시 - MessageReceived

        /// <summary>
        /// 메시지 수신시
        /// </summary>
        public event EventHandler<MessageReceivedEventArgs> MessageReceived;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 메시지 서버 URL
        /// </summary>
        private string messageServerURL = null;

        /// <summary>
        /// 메시지 주제
        /// </summary>
        private string messageSubject = null;

        /// <summary>
        /// 메시지 큐
        /// </summary>
        private string messageQueue = null;

        /// <summary>
        /// 연결
        /// </summary>
        private IConnection connection = null;

        /// <summary>
        /// 메시지 핸들러
        /// </summary>
        private EventHandler<MsgHandlerEventArgs> messageHandler = null;

        /// <summary>
        /// 비동기 구독 인터페이스
        /// </summary>
        private IAsyncSubscription subscription = null;

        /// <summary>
        /// 잠금 객체
        /// </summary>
        private object lockObject = new object();

        /// <summary>
        /// 작업 여부
        /// </summary>
        private bool isWorking = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - Receiver()

        /// <summary>
        /// 생성자
        /// </summary>
        public Receiver()
        {
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 초기화 하기 - Initialize(messageServerURL, messageSubject, messageQueue)

        /// <summary>
        /// 초기화 하기
        /// </summary>
        /// <param name="messageServerURL">메시지 서버 URL</param>
        /// <param name="messageSubject">메시지 주제</param>
        /// <param name="messageQueue">메시지 큐</param>
        public void Initialize(string messageServerURL, string messageSubject, string messageQueue)
        {
            this.messageServerURL = messageServerURL;
            this.messageSubject   = messageSubject;
            this.messageQueue     = messageQueue;

            Options options = ConnectionFactory.GetDefaultOptions();

            options.Url = this.messageServerURL;

            this.connection = new ConnectionFactory().CreateConnection(options);

            this.messageHandler = (sender, e) =>
            {
                lock(this.lockObject)
                {
                    FireMessageReceivedEvent(e.Message);
                }
            };
        }

        #endregion
        #region 서버 시작하기 - Start()

        /// <summary>
        /// 서버 시작하기
        /// </summary>
        public void Start()
        {
            if(this.isWorking)
            {
                return;
            }

            this.isWorking = true;

            try
            {
                this.subscription = this.connection.SubscribeAsync(this.messageSubject, this.messageQueue, this.messageHandler);
            }
            catch(Exception exception)
            {
                this.isWorking = false;

                throw exception;
            }
        }

        #endregion
        #region 서버 중단하기 - Stop()

        /// <summary>
        /// 서버 중단하기
        /// </summary>
        public void Stop()
        {
            if(!this.isWorking)
            {
                return;
            }

            this.isWorking = false;

            lock(this.lockObject)
            {
                if(this.subscription != null)
                {
                    this.subscription.Dispose();

                    this.connection.Dispose();
                }
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 메시지 수신시 이벤트 발생시키기 - FireMessageReceivedEvent(message)

        /// <summary>
        /// 메시지 수신시 이벤트 발생시키기
        /// </summary>
        /// <param name="message">메시지</param>
        protected void FireMessageReceivedEvent(Msg message)
        {
            MessageReceivedEventArgs e = new MessageReceivedEventArgs(message);

            MessageReceived?.Invoke(this, e);

            if(e.ReplyData != null)
            {
                message.ArrivalSubcription.Connection.Publish(message.Reply, e.ReplyData);
            }
        }

        #endregion
    }
}